/*

2clseq. Copyright Fahiem Bacchus 2000--2002

 A DPLL solver based on performing extensive reasoning with the binary
 clause subtheory. Pronounced "Two" "Clause" "Eq"

Version 2.0.

The software available at this site is copyright (c) 2000-2002 Fahiem
Bacchus. All right are reserved. Use of this software is permitted for
non-commercial research purposes, and it may be copied only for that
use. All copies must include this copyright message. This software and
any documentation and/or information supplied with it is distributed
on an as is basis. Fahiem Bacchus and the University of Toronto make
no warranties, express or implied, including but not limited to
implied warranties of merchantability and fitness for a particular
purpose, regarding the documentation, functions or performance of such
software, documentation and/or information.

@misc{Bacchus:2clseq,
	author  = { Fahiem Bacchus },
	year    = { 2002 },
	title   = { {2clseq} a {DPLL} solver employing extensive binary clause reasoning },
	note    = { Available from 
                    http://www.cs.toronto.edu/\~{}fbacchus/2clseq.html }
}

*/

#include "2clsEq.h"
#include "binaryClauses.h"
#include "naryClauses.h"
#include "resolve.h"

/********************************************************
  Resolve binary clauses
********************************************************/

void makeeq(Lit* x, Lit* l, DAG* dagnode);
EQS* eqbuf;

void resolvebcls(bcls *pbcls) {
  //Resolve a binary clause. Should be called only if upstack is empty.
  //I.e. there are no pending truth assignments.

  bclsResolved++;

  int l1=pbcls->l1;
  int l2=pbcls->l2;
  DAG* dagnode=pbcls->dagnode;
  DAG* dnode1;
  deallocbcls(pbcls);
  
#ifdef _TRACE
  printf("=>R2CL(%d[%c],%d[%c]).\n",
	 to_ff(l1), ff_tval((THELITS+l1)->val),
	 to_ff(l2), ff_tval((THELITS+l2)->val));
#endif
  
  Lit *NotL1=THELITS+negate(l1);
  Lit *L1=THELITS+l1;
  Lit *NotL2=THELITS+negate(l2);
  //  bucket *pbuc = L1->litHtable;
  //  bucket *notpbuc = NotL1->litHtable;
  //1. check for staleness.
  if(isFalse(l1) && isFalse(l2)) {
    setContradiction(getLitdagnode(NotL1),
		     allocDAGNODE_NORMAL(getLitdagnode(NotL2),dagnode));
    return;
  }
  //  if(isTrue(l1)||isTrue(l2)||HvalhaveBclause(pbuc,l2,RESOLVED)) {
  //  if(isTrue(l1)||isTrue(l2)||haveBclause(pbuc,l2)) {
  if(isTrue(l1)||isTrue(l2)||haveBclause(L1,l2)) {
    return;
  }

  //2. check for predecessors resolutions making this generate
  //   a unitprop
  if(isFalse(l1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(getLitdagnode(NotL1),dagnode));
    return;
  }
  //  if(DNodehaveBclause(notpbuc,l2,dnode1)) {
  if(DNodehaveBclause(NotL1,l2,dnode1)) {
    pushupstack(l2,allocDAGNODE_NORMAL(dagnode,dnode1));
    return;
  }
  if(isFalse(l2)) {
    pushupstack(l1,allocDAGNODE_NORMAL(getLitdagnode(NotL2),dagnode));
    return;
  }
  //  if(DNodehaveBclause(pbuc,negate(l2),dnode1)) {
  if(DNodehaveBclause(L1,negate(l2),dnode1)) {
    pushupstack(l1,allocDAGNODE_NORMAL(dagnode,dnode1));
    return;
  }

  //3. create all the new 2-clauses 
  BinLitCls* pb1;
  BinLitCls* pb2;  

  for(pb1=&(NotL1->twoClauses);pb1->pnext;pb1=pb1->pnext) {
    // pbuc=pb1->otherlit->litHtable;
    //notpbuc=negate(pb1->otherlit)->litHtable;
    for(pb2=&(NotL2->twoClauses);pb2->pnext;pb2=pb2->pnext) {
      if(pb1->otherlit==pb2->otherlit) {
	pushupstack(pb1->otherlit->num,
		    allocDAGNODE_NORMAL(dagnode,
					allocDAGNODE_NORMAL(pb1->dagnode,pb2->dagnode)));
	break;
      }
      //      else if(DNodehaveBclause(pbuc,negate(pb2->otherlit->num),dnode1)) {
      else if(DNodehaveBclause(pb1->otherlit,negate(pb2->otherlit->num),dnode1)) {
	pushupstack(pb1->otherlit->num,
		    allocDAGNODE_NORMAL(dagnode,
					allocDAGNODE_NORMAL(pb1->dagnode,
							    allocDAGNODE_NORMAL(pb2->dagnode,dnode1))));
	break;
      }
      //      else if(DNodehaveBclause(notpbuc,pb2->otherlit->num,dnode1))
      else if(DNodehaveBclause(negate(pb1->otherlit),pb2->otherlit->num,dnode1))
	pushupstack(pb2->otherlit->num,
		    allocDAGNODE_NORMAL(dagnode,
					allocDAGNODE_NORMAL(pb1->dagnode,
							    allocDAGNODE_NORMAL(pb2->dagnode,dnode1))));
      //      else if(!HvalhaveBclause(pbuc,pb2->otherlit->num,RESOLVED)) {
      //	  else if(!haveBclause(pbuc,pb2->otherlit->num)) {
      else if(!haveBclause(pb1->otherlit,pb2->otherlit->num)) {
	dnode1 = allocDAGNODE_NORMAL(dagnode,
				     allocDAGNODE_NORMAL(pb1->dagnode,pb2->dagnode));
	addBclause(pb1->otherlit,pb2->otherlit,dnode1);
	if(DOHYP)
	  pushclschkstack(pb1->otherlit,pb2->otherlit);
      }
      if(CONTRADICTION.val)
	return;
    }
  }
  
  if (DOEQS) {
    //test for generated equivalence.
    int topeqbuf=0;
    int i;
    
    //  pbuc = L1->litHtable;
    //  if(isUnvalued(L1) && HvalhaveBclause(negate(l1),negate(l2),RESOLVED)) {
    
    if(isUnvalued(L1) && haveBclause(negate(l1),negate(l2))) {
      for(pb1=NotL1->twoClauses.pnext;pb1->pnext;pb1=pb1->pnext) {
	//      if(DNodeHvalhaveBclause(pbuc,negate(pb1->otherlit->num),RESOLVED,dnode1)) {
	//	  if(DNodehaveBclause(pbuc,negate(pb1->otherlit->num),dnode1)) {
	if(DNodehaveBclause(L1,negate(pb1->otherlit->num),dnode1)) {
	  eqbuf[topeqbuf].L=pb1->otherlit;
	  //bclause negate(pb1->otherlit),L1 is reason pb1->otherlit
	  //can be replaced by L1.
	  eqbuf[topeqbuf].dagpos=dnode1;
	  //bclause NotL1,pb1->otherlit is reason negate(pb1->otherlit)
	  //can be replaced by NotL1.
	  eqbuf[topeqbuf].dagneg=pb1->dagnode;
	  topeqbuf++;
	}
      }
      
      for(i=0;i<topeqbuf;i++) {
	eqsfound++;
	makeeq(eqbuf[i].L,L1,eqbuf[i].dagpos);
	makeeq(negate(eqbuf[i].L),NotL1,eqbuf[i].dagneg);
      }
    }
  }
	
#ifdef _TRACE
  printf("<=R2CL(%d[%c],%d[%c]).\n",
	 to_ff(l1), ff_tval((THELITS+l1)->val),
	 to_ff(l2), ff_tval((THELITS+l2)->val));
#endif _TRACE
}

void makeeq(Lit* x, Lit* l, DAG* dagnode) {
  //rewrite all nary clauses containing x to contain l.
  //Then delink all b-clauses with x.
#ifdef _TRACE
  printf("=>makeeq(%d<=%d)\n",to_ff(l->num), to_ff(x->num));
  printLit(x);
  printLit(l);
#endif _TRACE
	
  if(isValued(x) || upstack->member(x->num)||upstack->member(negate(x)->num))
    return;
  markundoEq(x);

  x->equivlit=l;
  x->dagEqnode=dagnode;
  LitCls *plc;
	
  for(LitCls *pNclauses=x->nClauses.pnext;pNclauses->pnext;) {
    //must move pointer along inside of loop
    if(inClause(l,pNclauses->pcls)) {
      markundoClsLit(pNclauses->linkincls);
      delinkClsLit(pNclauses->linkincls);
      markundoClsDAG(pNclauses->pcls);
      pNclauses->pcls->dagnode=
	allocDAGNODE_NORMAL(pNclauses->pcls->dagnode,dagnode);
      if(pNclauses->pcls->len == 2)
	convertToBinaryClause(pNclauses->pcls);
      pNclauses=pNclauses->pnext;
    }
    else if(inClause(negate(l),pNclauses->pcls)) {
      plc=pNclauses;
      pNclauses=pNclauses->pnext;
      markundoClauseTrue(plc->pcls);
      delinkNClause(plc->pcls);
    }
    else {
      plc=pNclauses;
      pNclauses=pNclauses->pnext;
      markundoLitClsmove(plc,x);
      //delink from x.
      delinkLitCls(plc);
      //change to l
      plc->linkincls->plit=l;
      //relink to l
      relinkLitCls(plc, l);
      markundoClsDAG(plc->pcls);
      plc->pcls->dagnode=
	allocDAGNODE_NORMAL(plc->pcls->dagnode,dagnode);
    }
  }
  //Now unlink all binary clauses containing x.
  for(BinLitCls *pBclauses=x->twoClauses.pnext;pBclauses->pnext;
      pBclauses=pBclauses->pnext) 
    //for each true lit, we delink all of its binary clauses
    //in active lits.
    if(isUnvalued(pBclauses->otherlit)) {
      markundoBclause(pBclauses->otherlinks);
      delinkBclause(pBclauses->otherlinks);
    }
		
#ifdef _TRACE
  printLit(x);
  printLit(l);
  printf("<=makeeq(%d<-%d)\n",to_ff(x->num), to_ff(l->num));
#endif _TRACE
}

